#ifndef FUNC_FILE_HPP_GDAFVHTAVYHBTGFHNTIUYFHSYBSGFSDTGSVFAWASEHTYO
#define FUNC_FILE_HPP_GDAFVHTAVYHBTGFHNTIUYFHSYBSGFSDTGSVFAWASEHTYO

#include "sets/comparable_hash"
#include "robust_measure.hpp"
#include <string>
#include <random>
#include <algorithm>
#include <chrono>
#include <iostream>
#include <set>
#include <sstream>
#include <iostream>
#include <cmath>

template<typename C> struct needs_sorted_input : public std::false_type {};

template<typename T>
struct evaluate;

template<> struct evaluate<int>{ double operator()(int a){return std::abs(a);} };


const double startN = 10.0;
const double multiplier = std::pow( 10.0, 1.0/4.0 );
const int regenerate_period = 10;

template<int N = 50>
class record{
	public:
		using const_iterator = typename std::array<char, N>::const_iterator;

		char& operator[](int pos){ return str[pos]; }
		const char& operator[](int pos) const { return str[pos]; }
		bool operator<(const record& rec) const {
			for(int i = 0; i < N; ++i){
				if(str[i] < rec.str[i]){
					return true;				
				}
				else if(str[i] > rec.str[i]){
					return false;
				}
			}
			return false;
		}
		bool operator>(const record& rec) const {
			return rec < *this;
		}
		bool operator>=(const record& rec) const {
			return !(*this < rec);
		}

		bool operator==(const record& rec) const {
			for(int i = 0; i < N; ++i){
				if(str[i] != rec.str[i]){
					return false;
				}
			}
			return true;
		}
		bool operator!=(const record& rec) const {
			return !operator==(rec);		
		}


		const_iterator begin() const { return str.cbegin(); }
		const_iterator end() const { return str.cend(); }
		
	private:
		std::array<char, N> str;
};

template<int N> struct evaluate<record<N>>{
	double operator()(const record<N>& rec){
		double ret;
		for(auto a : rec){
			ret += std::abs(a);
		}
		return ret;
	} 
};

template<> struct evaluate<std::string>{
	double operator()(const std::string& str){
		double ret;
		for(auto a : str){
			ret += std::abs(a);
		}
		return ret;
	} 
};

namespace sets{
	template<int N>
	class comparable_hash<record<N>>{
		public:
			size_t operator()(const record<N>& rec){
				size_t ret = 0;
				for(int i = 0; i<N && i<8; ++i){
					int a = rec[i] - std::numeric_limits<char>::min();
					assert(a > 0);
					ret |= static_cast<size_t>(a) << ((7-i)*8);
				}
				return ret;
			}
	};
}

template<typename T>
std::ostream& operator<<(std::ostream& out, const std::unique_ptr<T>& ptr){
	return out << ptr.get();
}

template<int size>
std::ostream& operator<<(std::ostream& out, const record<size>& rec){
	for(char c : rec){
		out << c;
	}
	return out;
}

auto get_random_engine(){
	std::random_device rd{};
	return std::mt19937{rd()};
}

template<typename T>
struct element_source;

template<>
struct element_source<double>{
	element_source() : engine(std::random_device{}()), urd{} {}
	double operator()(){ return urd(engine); }
	private:
		std::mt19937 engine;
		std::uniform_real_distribution<double> urd;
};

template<>
struct element_source<int>{
	element_source() : engine(std::random_device{}()), uid{} {}
	int operator()(){ return uid(engine); }
	private:
		std::mt19937 engine;
		std::uniform_int_distribution<int> uid;
};

template<>
struct element_source<std::string>{
	element_source() : engine(std::random_device{}()), uid{} {}
	std::string operator()(){ 
		std::string ret;
		for(int i = 0; i < 50; ++i){
			ret += uid(engine);
		}
		return ret;
	}
	private:
		std::mt19937 engine;
		std::uniform_int_distribution<char> uid;
};

template<int N>
struct element_source<record<N>>{
	element_source() : engine(std::random_device{}()), uid{} {}
	record<N> operator()(){
		record<N> ret; 
		for(int i = 0; i < 50; ++i){
			ret[i] = uid(engine);
		}
		return ret;
	}
	private:
		std::mt19937 engine;
		std::uniform_int_distribution<char> uid;
};

template<typename T>
struct element_source<std::unique_ptr<T>>{
	std::unique_ptr<T> operator()(){ 
		return move(std::unique_ptr<T>(new T{source()}));
	}
	private:
		element_source<T> source;
};

template<typename test_t>
void perform_test(
		std::function<void()> initialize,
		std::function<void()> pre_test, std::function<void()> post_test, 
		std::function<void()> test_setup, std::function<void()> test_teardown,
		test_t test, std::function<double()> N){
	double last_measure = 0.0;
	double last_measuring_time = 0.0;
	auto now = std::chrono::high_resolution_clock::now;

	initialize();
	while( last_measure < 10000000 && last_measuring_time < 500000000){
		auto start = now();
		pre_test();
		last_measure = robust_measure([&]{
			test_setup();
			auto start = now();
			test();
			auto dur = now() - start;
			test_teardown();
			return std::chrono::duration_cast<std::chrono::nanoseconds>(dur).count();
		}, .1);
		post_test();
		last_measuring_time = std::chrono::duration_cast<std::chrono::nanoseconds>(now() - start).count();
	}

	last_measure = 0;
	initialize();
	while( last_measure < 100000000 && N() < 10000000.1){
		pre_test();
		last_measure = robust_measure([&]{
			test_setup();
			auto start = now();
			test();
			auto dur = now() - start;
			test_teardown();
			return (double)std::chrono::duration_cast<std::chrono::nanoseconds>(dur).count();
		}, .001, std::max(50, int(500000/N()) ));
		std::cout << "N: " << N() << '\t' << last_measure/N() << std::endl;
		post_test();
	}
}

template<typename C, typename T>
void growth_insert_ascending(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	C cont;
	std::vector<T> numbers;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear(); 
		std::generate_n(std::back_inserter(numbers), N, random);
		std::sort(numbers.begin(), numbers.end());
	};

	auto pre_test = [&]{ regenerate(); };

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();
		}
		cont.clear();
	};
	auto test = [&]{
		for(T& num : numbers){
			cont.insert(num);
		}
	};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, []{}, test, [&]{return N;});
	std::cout << std::endl;
}

template<typename C, typename T>
void growth_insert_descending(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	C cont;
	std::vector<T> numbers;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear(); 
		std::generate_n(std::back_inserter(numbers), N, random);
		std::sort(numbers.begin(), numbers.end(), std::greater<T>{});
	};

	auto pre_test = [&]{ regenerate(); };

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();
		}
		cont.clear();
	};
	auto test = [&]{
		for(T& num : numbers){
			cont.insert(num);
		}
	};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, []{}, test, [&]{return N;});
	std::cout << std::endl;
}

template<typename C, typename T>
void growth_insert_middle(std::string func, std::string structure, std::string type ){

	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	C cont;
	std::vector<T> numbers;
	std::vector<T> generated;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear();
		generated.clear();
		std::generate_n(std::back_inserter(generated), N, random);
		std::sort(generated.begin(), generated.end());

		auto it1 = generated.begin();
		auto it2 = generated.end();
		while(it1 < it2){
			numbers.push_back(std::move(*(it1++)));
			if(it1 < it2){
				numbers.push_back(std::move(*(--it2)));
			}
		}
	};

	auto pre_test = [&]{ regenerate(); };

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();
		}
		cont.clear();
	};
	auto test = [&]{
		for(T& num : numbers){
			cont.insert(num);
		}
	};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, []{}, test, [&]{return N;});
	std::cout << std::endl;
}

template<typename C, typename T>
void growth_insert_random(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	C cont;
	std::vector<T> numbers;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear();
		std::generate_n(std::back_inserter(numbers), N, random);
		std::shuffle(numbers.begin(), numbers.end(), engine);
	};
	
	auto pre_test = [&]{ regenerate(); };

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();
		}
		cont.clear();
	};
	auto test = [&]{
		for(T& num : numbers){
			cont.insert(num);
		}
	};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, []{}, test, [&]{return N;});
	std::cout << std::endl;
}

template<typename C, typename T>
void shrink_erase_random(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	C cont;
	std::vector<T> numbers;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear();
		std::generate_n(std::back_inserter(numbers), N, random);
		std::shuffle(numbers.begin(), numbers.end(), engine);
	};

	auto pre_test = [&]{
		regenerate();
	};

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();		
		}
		for(const T& num : numbers){
			cont.insert(num);
		}
		std::shuffle(numbers.begin(), numbers.end(), engine);
	};
	auto test = [&]{
		for(T& num : numbers){
			cont.erase(num);
		}	
	};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, []{}, test, [&]{return N;});
	std::cout << std::endl;
}

template<typename C, typename T>
void forward_iteration(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	auto evaluator = evaluate<T>();
	C cont;
	std::vector<T> numbers;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear();
		std::generate_n(std::back_inserter(numbers), N, random);
		if(needs_sorted_input<C>::value){
			std::sort(numbers.begin(), numbers.end());	
		}
		else{
			std::shuffle(numbers.begin(), numbers.end(), engine);
		}
		cont.clear();
		for(T& num : numbers){
			cont.insert(std::move(num));
		}
	};

	auto pre_test = [&]{
		regenerate();
	};

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();		
		}
	};
	auto test = [&]{
		for(const auto& item : cont)
		total += evaluator(item);
	};
	auto test_teardown = [&]{ if(total < 10.0){ std::abort(); }};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, test_teardown, test, [&]{ return N; });
	std::cout << std::endl;
}

template<typename C, typename T>
void correctness_test(std::string func, std::string structure, std::string type){
	
	C cont;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	
	std::set<T> correct;	

	int N = 10000;
	int k = 1;

	for(int i = 0; i < N; ++i){
		T element = random();

		if(correct.find(element) == correct.end() && cont.find(element) != cont.end()){
			throw std::logic_error("Found element before insertion");
		}

		auto ret1 = cont.insert(element);
		auto ret2 = correct.insert(element);

		if(cont.size() != correct.size()){
			std::cerr << cont.size() << ", " << correct.size() << std::endl;
			throw std::logic_error("size mismatch!");
		}

		if(ret1.second != ret2.second){
			std::stringstream ss;
			ss << std::boolalpha;
			ss << structure << " returned " << ret1.second << ", ";
			ss << "set returned " << ret2.second << std::endl;
			throw std::logic_error(ss.str());		
		}

		auto it = cont.find(element);
		if(it == cont.end()){
			throw std::logic_error("Could not find element!");
		}
		if(*it != element){
			throw std::logic_error("Find returned wrong iterator!");
		}
		
	
		if(*ret1.first != element){
			throw std::logic_error(structure + " iterator doesn't match proxy!");
		}

		auto ret3 = cont.insert(element);		

		if(cont.size() != correct.size()){
			std::cerr << cont.size() << ", " << correct.size() << std::endl;
			throw std::logic_error("size changed after duplicate insertion!");
		}

		if(ret3.second){
			throw std::logic_error("Returned true from duplicate insertion");		
		}

		if(i >= k){
			if(!std::equal(cont.begin(), cont.end(), correct.begin())){
				throw std::logic_error(structure + " does not match set!");
			}
			if(!std::equal(correct.begin(), correct.end(), cont.begin())){
				throw std::logic_error("set does not match " + structure + "!");
			}
			std::cout << "i: " << i << std::endl;
			k *= 2;
		}
	}

	if(cont.size() != correct.size()){
		throw std::logic_error("size mismatch at end!");
	}

	if(!std::equal(cont.begin(), cont.end(), correct.begin())){
		throw std::logic_error(structure + " does not match set at end!");
	}
	if(!std::equal(correct.begin(), correct.end(), cont.begin())){
		throw std::logic_error("set does not match " + structure + " at end!");
	}
	std::cout << "All insertion tests passed :)" << std::endl;

	std::vector<T> vec(correct.begin(), correct.end());
	std::shuffle(vec.begin(), vec.end(), engine);
	
	k = 1;
	int i = 0;

	for(auto proxy : vec){
		correct.erase(proxy);
		auto ret2 = cont.erase(proxy);
		if(ret2 != 1){
			throw std::logic_error("Error in return value of erase");
		}

		if(cont.size() != correct.size()){
			throw std::logic_error("Size mismatch after deletion!");
		}

		auto it = cont.find(proxy);
		if(it != cont.end()){
			throw std::logic_error("Proxy found after deletion!");
		}
	
		if(i >= k || proxy == vec.back()){
			if(std::distance(cont.begin(), cont.end()) != cont.size()){
				throw std::logic_error("Iterator distance does not match size");
			}
			if(!std::equal(cont.begin(), cont.end(), correct.begin())){
				throw std::logic_error(structure + " does not match set!");
			}
			if(!std::equal(correct.begin(), correct.end(), cont.begin())){
				throw std::logic_error("set does not match " + structure + "!");
			}
			std::cout << "i: " << i << std::endl;
			k *= 2;
		}
		i++;
	}
	std::cout << "All deletion tests passed :)" << std::endl;
}

template<typename C, typename T>
void search_random(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	C cont;
	std::vector<T> numbers;
	std::vector<T> queries;
	int finds = 0;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear();
		std::generate_n(std::back_inserter(numbers), long(N), random);
		if(needs_sorted_input<C>::value){
			std::sort(numbers.begin(), numbers.end());
		}
		else{
			std::shuffle(numbers.begin(), numbers.end(), engine);
		}
		cont.clear();
		for(T& num : numbers){
			cont.insert(num);
		}
		std::shuffle(numbers.begin(), numbers.end(), engine);
	};

	auto pre_test = [&]{
		regenerate();
	};

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();		
		}
	};
	auto test = [&]{
		for(auto& query : numbers){
			if(cont.find(query) != cont.end()){
				++finds;
			}
		}
	};
	
	auto test_teardown = [&]{ 
		if(finds > long(N)){
			std::cout << "waa!";		
		}
		finds = 0;
	};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, test_teardown, test, [&]{ return N; });
	std::cout << std::endl;
}

template<typename C, typename T>
void search_ascending(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	C cont;
	std::vector<T> numbers;
	std::vector<T> queries;
	int finds = 0;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear();
		std::generate_n(std::back_inserter(numbers), long(N), random);
		if(needs_sorted_input<C>::value){
			std::sort(numbers.begin(), numbers.end());
		}
		else{
			std::shuffle(numbers.begin(), numbers.end(), engine);
		}
		cont.clear();
		for(T& num : numbers){
			cont.insert(num);
		}
		if(!needs_sorted_input<C>::value){
			std::sort(numbers.begin(), numbers.end());
		}
	};

	auto pre_test = [&]{
		regenerate();
	};

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();		
		}
	};
	auto test = [&]{
		for(auto& query : numbers){
			if(cont.find(query) != cont.end()){
				++finds;
			}
		}
	};
	
	auto test_teardown = [&]{ 
		if(finds > long(N)){
			std::cout << "waa!";		
		}
		finds = 0;
	};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, test_teardown, test, [&]{ return N; });
	std::cout << std::endl;
}

template<typename C, typename T>
void indel_random(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	C cont;
	std::vector<T> numbers1;
	std::vector<T> numbers2;
	int finds = 0;

	auto initialize = [&]{ N = startN; total = 0; };

	auto pre_test = [&]{
		numbers1.clear();
		std::generate_n(std::back_inserter(numbers1), N, random);
		std::shuffle(numbers1.begin(), numbers1.end(), engine);

		numbers2.clear();
		std::generate_n(std::back_inserter(numbers2), N, random);
		std::shuffle(numbers2.begin(), numbers2.end(), engine);
	};

	auto test_setup = [&]{
		cont.clear();
		for(auto& num : numbers1){
			cont.insert(num);	
		}
		std::shuffle(numbers1.begin(), numbers1.end(), engine);
		std::shuffle(numbers2.begin(), numbers2.end(), engine);
	};
	auto test = [&]{
		for(int i = 0; i < N; ++i){
			cont.insert(numbers2[i]);
			cont.erase(numbers1[i]);
		}
	};
	
	auto test_teardown = [&]{};

	auto post_test = [&]{ N *= multiplier; };

	perform_test(initialize, pre_test, post_test, test_setup, test_teardown, test, [&]{return N;});
	std::cout << std::endl;
}

template<typename C, typename T>
void accumulation(std::string func, std::string structure, std::string type ){
	
	double N;
	double total;
	auto engine = get_random_engine();
	auto random = element_source<T>();
	auto evaluator = evaluate<T>();
	C cont;
	std::vector<T> numbers;

	auto initialize = [&]{ N = startN; total = 0; };

	auto regenerate = [&]{
		numbers.clear();
		std::generate_n(std::back_inserter(numbers), N, random);
		if(needs_sorted_input<C>::value){
			std::sort(numbers.begin(), numbers.end());	
		}
		else{
			std::shuffle(numbers.begin(), numbers.end(), engine);
		}
		cont.clear();
		for(T& num : numbers){
			cont.insert(std::move(num));
		}
	};

	auto pre_test = [&]{
		regenerate();
	};

	auto test_setup = [&]{
		if(engine() % regenerate_period == 0){
			regenerate();		
		}
	};
	auto test = [&]{
		total += cont.accumulate(evaluator, 0.0);
	};

	auto post_test = [&]{ N *= multiplier; };
	auto test_teardown = [&]{
		if(total < 10.0){
			std::abort();
		}
	};

	auto getN = [&]{return N;};

	perform_test(initialize, pre_test, post_test, test_setup, test_teardown, test, getN);
	std::cout << std::endl;
}

#endif
